#include <iostream>
#include <thread>
#include <chrono>
#include <string>
#include <nanomsg/nn.h>
#include <nanomsg/reqrep.h>
#include "rpcInterface.pb.h"
using namespace std;

//不同cpu下int在内存里的顺序不一样，不能莽memcpy
inline void int_to_bytes(int a, uint8_t *address) {
    address[0] = a>>24; address[1] = a>>16;
    address[2] = a>>8;  address[3] = a;
}


// @class rpcChannel
// 用户调用rpc以后，系统会自动调这个。
class rpcChannel : public google::protobuf::RpcChannel {
public:
    int fd;
    rpcChannel(int _fd): fd(_fd){};
    //function calling will be processing here..
    void CallMethod(const google::protobuf::MethodDescriptor* method,
                    google::protobuf::RpcController* controller,
                    const google::protobuf::Message* request,
                    google::protobuf::Message* response,
                    google::protobuf::Closure* done) {        
        char buffer[1024];

        // 取调用的 function_id，填充数据包前4字节。
        uint8_t opcode[4];
        int_to_bytes(method->index(), opcode);
        memcpy(buffer, opcode, 4);
        
        // 把request填充到剩余部分。
        const string &tmp = request->SerializeAsString();
        memcpy(buffer+4, tmp.c_str(), 4+tmp.length());

        //通过nanomsg把这个包送出去 4字节函数号+其余request
        cout<<"[nanomsg] pre-send"<<endl;
        int rc=nn_send(fd, buffer, 4+tmp.length(), 0);
        assert(rc == 4+tmp.length());
        cout<<"[nanomsg] sent & pre-recv"<<endl;
        rc = nn_recv(fd, buffer, sizeof(buffer), 0);

        //把返回的包给response
        response->ParseFromString( std::string(buffer,rc) );
    };
};


// rpc客户端，先初始化nanomsg，注册中间件rpcChannel，
// 然后以后调rpc方法时，系统会自动调rpcChannel中间件（上面的class），
// 我们在里面写了nanomsg通信代码，从而把调用送给rpc服务端。
int main() {
    cout<<"** Client Running **"<<endl;
    int fd = nn_socket(AF_SP, NN_REQ);
    assert(fd >= 0);
    int ret = nn_connect(fd, "ipc://rpc_address.proto");
    assert(ret >= 0);
    this_thread::sleep_for( chrono::microseconds(200) ); //waiting nanomsg to complete the link.

    //给一个protobuf.service 的service处理函数，以后调rpc函数都从stub开始。
    myrpc::rpcService_Stub stub( new rpcChannel(fd) );

    //try to calling rpc function add.
    cout<<"Calling RPC.add(33,66)"<<endl;
    myrpc::rpcRequest ask; ask.set_a(33); ask.set_b(66);
    myrpc::rpcResponse ans;
    stub.add(nullptr, &ask, &ans, nullptr);
    cout<<ans.result()<<endl;
    cout<<"** Client Exit **"<<endl;
    return 0;
}